import ExpoModulesCore
import Foundation

public class ExpoIcloudStorageModule: Module {
    private func uploadFile(destinationPath: String, filePath: String, documentsURL: URL, completionHandler: @escaping (Result<String, Error>) -> Void) {
        let destinationURL = documentsURL.appendingPathComponent(destinationPath)

        if FileManager.default.fileExists(atPath: destinationURL.path) {
            print("File exist")
            completionHandler(.success(destinationURL.path))
            return
        }

        do {
            try FileManager.default.createDirectory(at: destinationURL.deletingLastPathComponent(), withIntermediateDirectories: true, attributes: nil)
        } catch let error as NSError {
            print("Error creating directory: \(error.localizedDescription)")
            completionHandler(.failure(error))
            return
        }

        let filePathWithoutPrefix = filePath.replacingOccurrences(of: "file://", with: "")
        let newFileURL = documentsURL.appendingPathComponent(destinationPath, isDirectory: false)

        DispatchQueue.global(qos: .background).async {
            var uploadError: Error?

            do {
                try FileManager.default.copyItem(atPath: filePathWithoutPrefix, toPath: newFileURL.path)
            } catch let error as NSError {
                print("Error copying file: \(error.localizedDescription)")
                uploadError = error
            }

            DispatchQueue.main.async {
                if let error = uploadError {
                    completionHandler(.failure(error))
                } else {
                    completionHandler(.success(destinationURL.path))
                }
            }
        }
    }

    private func downloadFile(path: String, destinationDir: String, completionHandler: @escaping (Result<String, Error>) -> Void) {
        let fileURL = URL(fileURLWithPath: path)
        let destinationDirURL = URL(fileURLWithPath: destinationDir)
        let fileManager = FileManager.default
        let lastPathComponent = fileURL.lastPathComponent
        let destinationURL = destinationDirURL.appendingPathComponent(fileManager.displayName(atPath: fileURL.path))

        func copyAndFinish(sourceURL: URL) {
            do {
                if fileManager.fileExists(atPath: destinationURL.path) {
                    completionHandler(.success(destinationURL.path))
                    return
                }
                try fileManager.copyItem(at: sourceURL, to: destinationURL)
                completionHandler(.success(destinationURL.path))
            } catch {
                print("Error copying file: \(error)")
                completionHandler(.failure(error))
            }
        }

        if lastPathComponent.contains(".icloud") {
            let query = NSMetadataQuery()
            query.predicate = NSPredicate(format: "%K == %@", NSMetadataItemPathKey, fileURL.path)
            query.searchScopes = [NSMetadataQueryUbiquitousDocumentsScope]
            
            NotificationCenter.default.addObserver(forName: .NSMetadataQueryDidUpdate, object: query, queue: nil) { notification in
                for i in 0..<query.resultCount {
                    if let item = query.result(at: i) as? NSMetadataItem {
                        let downloadingStatus = item.value(forAttribute: NSMetadataUbiquitousItemDownloadingStatusKey) as! String

                        if downloadingStatus == URLUbiquitousItemDownloadingStatus.current.rawValue {
                            // File is downloaded
                            query.stop()
                            NotificationCenter.default.removeObserver(self, name: .NSMetadataQueryDidUpdate, object: query)
                            DispatchQueue.main.async {
                                copyAndFinish(sourceURL: fileURL)
                            }
                        }
                    }
                }
            }

            DispatchQueue.global(qos: .userInitiated).async {
                do {
                    try fileManager.startDownloadingUbiquitousItem(at: fileURL)
                } catch {
                    print("Error starting download: \(error)")
                }
                query.start()
            }
        } else {
            copyAndFinish(sourceURL: fileURL)
        }
    }



    public func definition() -> ModuleDefinition {
        Name("ExpoIcloudStorage")
        Events("onUploadFilesAsyncProgress", "onDownloadFilesAsyncProgress")
        Constants([
            "defaultICloudContainerPath": FileManager.default.url(forUbiquityContainerIdentifier: nil)?.path ?? nil,
        ])

        Function("getDefaultPathAsync") {
            FileManager.default.url(forUbiquityContainerIdentifier: nil)?.path ?? nil
        }

        AsyncFunction("isICloudAvailableAsync") { (promise: Promise) in
            let isICloudAvailable = FileManager.default.ubiquityIdentityToken != nil
            promise.resolve(isICloudAvailable)
        }

        AsyncFunction("isExistAsync") { (path: String, isDirectory: Bool, promise: Promise) in
            guard let documentsURL = FileManager.default.url(forUbiquityContainerIdentifier: nil)?.appendingPathComponent("Documents") else {
                promise.reject(NSError(domain: "expo-icloud-storage", code: 0, userInfo: [NSLocalizedDescriptionKey: "iCloud documents directory not found"]))
                return
            }
            let fileURL = documentsURL.appendingPathComponent(path, isDirectory: isDirectory)
            promise.resolve(FileManager.default.fileExists(atPath: fileURL.path))
        }

        AsyncFunction("createDirAsync") { (path: String, promise: Promise) in
            guard let documentsURL = FileManager.default.url(forUbiquityContainerIdentifier: nil)?.appendingPathComponent("Documents") else {
                promise.reject(NSError(domain: "expo-icloud-storage", code: 0, userInfo: [NSLocalizedDescriptionKey: "iCloud documents directory not found"]))
                return
            }

            let directoryURL = documentsURL.appendingPathComponent(path, isDirectory: true)

            do {
                try FileManager.default.createDirectory(at: directoryURL, withIntermediateDirectories: true, attributes: nil)
                promise.resolve(true)
            } catch let error as NSError {
                print("Error creating directory: \(error.localizedDescription)")
                promise.reject(error)
            }
        }

        AsyncFunction("unlinkAsync") { (path: String, promise: Promise) in
            let fileURL = URL(fileURLWithPath: path)

            guard let iCloudContainer = FileManager.default.url(forUbiquityContainerIdentifier: nil),
                  let iCloudDocumentsURL = FileManager.default.url(forUbiquityContainerIdentifier: nil)?.appendingPathComponent("Documents"),
                  fileURL.path.hasPrefix(iCloudContainer.path) else {
                promise.reject(NSError(domain: "expo-icloud-storage", code: 0, userInfo: [NSLocalizedDescriptionKey: "Invalid iCloud path"]))
                return
            }

            do {
                try FileManager.default.removeItem(at: fileURL)
                promise.resolve(true)
            } catch let error as NSError {
                print("Error removing item: \(error.localizedDescription)")
                promise.reject(error)
            }
        }

        AsyncFunction("readDirAsync") { (path: String, options: [String: Bool], promise: Promise) in

            guard let documentsURL = FileManager.default.url(forUbiquityContainerIdentifier: nil)?.appendingPathComponent("Documents") else {
                promise.reject(NSError(domain: "expo-icloud-storage", code: 0, userInfo: [NSLocalizedDescriptionKey: "iCloud documents directory not found"]))
                return
            }
            let directoryURL = documentsURL.appendingPathComponent(path, isDirectory: true)
            guard let contents = try? FileManager.default.contentsOfDirectory(at: directoryURL, includingPropertiesForKeys: nil) else {
                promise.resolve([])
                return
            }

            let isFullPath = options["isFullPath"] ?? true
            print(isFullPath)
            if isFullPath {
                let fullPaths = contents.map { $0.path }
                promise.resolve(fullPaths)
            } else {
                let fileNames = contents.map { $0.lastPathComponent }
                promise.resolve(fileNames)
            }
        }

        AsyncFunction("downloadFileAsync") { (path: String, destinationDir: String, promise: Promise) in
            guard (FileManager.default.url(forUbiquityContainerIdentifier: nil)?.appendingPathComponent("Documents")) != nil else {
                promise.reject(NSError(domain: "expo-icloud-storage", code: 0, userInfo: [NSLocalizedDescriptionKey: "iCloud documents directory not found"]))
                return
            }
            let semaphore = DispatchSemaphore(value: 0)

            downloadFile(path: path, destinationDir: destinationDir) { result in
                switch result {
                case let .success(destinationPath):
                    promise.resolve(destinationPath)
                case let .failure(error):
                    promise.reject(error)
                }
                semaphore.signal()
            }

            semaphore.wait()
        }

        AsyncFunction("downloadFilesAsync") { (paths: [String], destinationDir: String, promise: Promise) in
            guard (FileManager.default.url(forUbiquityContainerIdentifier: nil)?.appendingPathComponent("Documents")) != nil else {
                promise.reject(NSError(domain: "expo-icloud-storage", code: 0, userInfo: [NSLocalizedDescriptionKey: "iCloud documents directory not found"]))
                return
            }

            let totalFiles = paths.count
            var filesProcessed = 0
            var results: [Result<String, Error>] = []
            self.sendEvent("onDownloadFilesAsyncProgress", ["value": 0])
            for path in paths {
                downloadFile(path: path, destinationDir: destinationDir) { result in
                    filesProcessed += 1
                    results.append(result)

                    let overallProgress = Int((Double(filesProcessed) / Double(totalFiles)) * 100)
                    self.sendEvent("onDownloadFilesAsyncProgress", ["value": overallProgress])

                    if filesProcessed == totalFiles {
                        promise.resolve(results.map { result -> [String: Any] in
                            switch result {
                            case let .success(destinationPath):
                                return ["success": true, "path": destinationPath]
                            case let .failure(error):
                                return ["success": false, "error": error.localizedDescription]
                            }
                        })
                    }
                }
            }
        }

        AsyncFunction("uploadFileAsync") { (destinationPath: String, filePath: String, promise: Promise) in
            guard let documentsURL = FileManager.default.url(forUbiquityContainerIdentifier: nil)?.appendingPathComponent("Documents") else {
                promise.reject(NSError(domain: "expo-icloud-storage", code: 0, userInfo: [NSLocalizedDescriptionKey: "iCloud documents directory not found"]))
                return
            }

            uploadFile(destinationPath: destinationPath, filePath: filePath, documentsURL: documentsURL) { result in
                switch result {
                case let .success(destinationPath):
                    promise.resolve(destinationPath)
                case let .failure(error):
                    promise.reject(error)
                }
            }
        }

        AsyncFunction("uploadFilesAsync") { (destinationDirectory: String, filePaths: [String], promise: Promise) in
            guard let documentsURL = FileManager.default.url(forUbiquityContainerIdentifier: nil)?.appendingPathComponent("Documents") else {
                promise.reject(NSError(domain: "expo-icloud-storage", code: 0, userInfo: [NSLocalizedDescriptionKey: "iCloud documents directory not found"]))
                return
            }
            let totalFiles = filePaths.count
            var filesProcessed = 0
            var results: [Result<String, Error>] = []

            for (_, filePath) in filePaths.enumerated() {
                let filePathWithoutPrefix = filePath.replacingOccurrences(of: "file://", with: "")
                let fileURL = URL(fileURLWithPath: filePathWithoutPrefix)
                let destinationPath = destinationDirectory + "/" + fileURL.lastPathComponent

                uploadFile(destinationPath: destinationPath, filePath: filePath, documentsURL: documentsURL) { result in
                    filesProcessed += 1
                    results.append(result)

                    let overallProgress = Int((Double(filesProcessed) / Double(totalFiles)) * 100)
                    self.sendEvent("onUploadFilesAsyncProgress", ["value": overallProgress])

                    if filesProcessed == totalFiles {
                        promise.resolve(results.map { result -> [String: Any] in
                            switch result {
                            case let .success(destinationPath):
                                return ["success": true, "path": destinationPath]
                            case let .failure(error):
                                return ["success": false, "error": error.localizedDescription]
                            }
                        })
                    }
                }
            }
        }
    }
}
